-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im  = ui_imgui

local C = {}

C.name = 'Vehicle Road Tracking'
C.description = 'Checks if a vehicle is driving on the road.'
C.color = im.ImVec4(1, 1, 0, 0.75)
C.category = 'repeat_instant'

C.pinSchema = {
  {dir = 'in', type = 'flow', name = 'flow'},
  {dir = 'in', type = 'number', name = 'vehId', description = 'Vehicle id to get information from; if none given, uses the player vehicle.'},
  {dir = 'in', type = 'table', name = 'aiPath', tableType = 'navPathData', description = '(Optional) Complete navgraph path to use for checking with.'},
  {dir = 'in', type = 'number', name = 'strictness', default = 1, description = 'Minimum number of vehicle corners that need to touch the road.'},

  {dir = 'out', type = 'flow', name = 'flow'},
  {dir = 'out', type = 'flow', name = 'onRoad', description = 'Sends flow if the vehicle is on a road.'},
  {dir = 'out', type = 'flow', name = 'offRoad', description = 'Sends flow if the vehicle is not on a road.'},
  {dir = 'out', type = 'number', name = 'distance', description = 'Distance past the edge of the nearest road (negative if on road).'},
  {dir = 'out', type = 'number', name = 'dirValue', description = 'Directional value of the nearest road (from 0 to 1)'}
}

function C:init()
  self.wpDict = {}
end

function C:_executionStopped()
  self.wpDict = {}
end

function C:work()
  local vehId = self.pinIn.vehId.value or be:getPlayerVehicleID(0)
  local aiPath = self.pinIn.aiPath.value
  local mapNodes = map.getMap().nodes
  local strictness = clamp(self.pinIn.strictness.value or 0, 1, 4)
  local n1, n2
  local onRoad = false

  if not next(self.wpDict) and aiPath then
    self.wpDict = tableValuesAsLookupDict(aiPath)
  end

  local obj = be:getObjectByID(vehId)
  if obj then
    local vehBB = obj:getSpawnWorldOOBB()
    local pos = vehBB:getCenter()
    local dirVec = obj:getDirectionVector()

    n1, n2 = map.findClosestRoad(pos)
    if n1 then
      if next(self.wpDict) then
        if self.wpDict[n1] and not self.wpDict[n2] then
          local idx = arrayFindValueIndex(aiPath, n1)
          local a, b = aiPath[idx - 1], aiPath[idx + 1]
          if (a or not b) and 1 / pos:xnormOnLine(mapNodes[n1].pos, mapNodes[a].pos) >= 1 then
            n2 = a
          else
            n2 = b
          end
        elseif self.wpDict[n2] and not self.wpDict[n1] then
          local idx = arrayFindValueIndex(aiPath, n2)
          local a, b = aiPath[idx - 1], aiPath[idx + 1]
          if (a or not b) and 1 / pos:xnormOnLine(mapNodes[n2].pos, mapNodes[a].pos) >= 1 then
            n1 = a
          else
            n1 = b
          end
        elseif not self.wpDict[n1] and not self.wpDict[n2] then
          n1, n2 = nil, nil
        end
      end

      if n1 and n2 then
        local p1, p2 = mapNodes[n1].pos, mapNodes[n2].pos
        if (p2 - p1):dot(dirVec) < 0 then
          n1, n2 = n2, n1
          p1, p2 = p2, p1
        end

        self.dirValue = math.abs(dirVec:dot((p2 - p1):normalized()))
        self.distance = math.huge
        local validCount = 0

        for i, v in ipairs({0, 3, 7, 4}) do
          local cPos = vehBB:getPoint(v)
          local xnorm = clamp(cPos:xnormOnLine(p1, p2), 0, 1)
          local roadPos = linePointFromXnorm(p1, p2, xnorm)
          local radius = lerp(mapNodes[n1].radius, mapNodes[n2].radius, xnorm)
          local normal = (p2 - p1):normalized():cross(mapNodes[n1].normal)
          local side = sign2((cPos - roadPos):dot(normal))
          local edge = roadPos + normal * radius * side
          local edgeSide = sign2((cPos - edge):dot(normal * side))

          self.distance = math.min(self.distance, cPos:distance(edge) * edgeSide)

          if edgeSide == -1 then
            validCount = validCount + 1
          end
        end

        if validCount >= strictness then
          onRoad = true
        end
      end
    end
  end

  if not obj or not n1 then
    self.distance = 0
    self.dirValue = 0
  end

  self.pinOut.onRoad.value = onRoad
  self.pinOut.offRoad.value = not onRoad
  self.pinOut.distance.value = self.distance
  self.pinOut.dirValue.value = self.dirValue
  self.pinOut.flow.value = self.pinIn.flow.value
end

return _flowgraph_createNode(C)